// Use PAM for HTTP basic authentication

#include <apache2/ap_provider.h>
#include <apache2/httpd.h>
#include <apache2/http_config.h>
#include <apache2/http_core.h>
#include <apache2/http_log.h>
#include <apache2/mod_auth.h>
#include <security/pam_appl.h>

#ifdef APLOG_USE_MODULE
APLOG_USE_MODULE(authn_pam);
#endif

typedef struct {
	unsigned fail_delay;
	const char *service_name;
} dir_config;

module AP_MODULE_DECLARE_DATA authn_pam_module;

static authn_status check_password(request_rec *r, const char *user,
	const char *password);
static int conv(int num_msg, const struct pam_message **msg,
	struct pam_response **resp, void *appdata_ptr);
static void *create_dir_config(apr_pool_t *p, char *dirspec);
static authn_status get_realm_hash(request_rec *r, const char *user,
	const char *realm, char **rethash);
static void register_hooks(apr_pool_t *p);

static const command_rec command_table[] = {
	AP_INIT_TAKE1("AuthPAMFailDelay", ap_set_int_slot,
		(void *) APR_OFFSETOF(dir_config, fail_delay), OR_AUTHCFG,
		"Number of microseconds to wait after an authentication "
		"failure (default is 0)"),
	AP_INIT_TAKE1("AuthPAMServiceName", ap_set_string_slot,
		(void *) APR_OFFSETOF(dir_config, service_name), OR_AUTHCFG,
		"PAM service name (default is apache2)"),
	{ .name = NULL }
};

module AP_MODULE_DECLARE_DATA authn_pam_module = {
	STANDARD20_MODULE_STUFF,
	create_dir_config,
	NULL,
	NULL,
	NULL,
	command_table,
	register_hooks
};

static const authn_provider authn_pam_provider = { check_password,
	get_realm_hash };

static authn_status check_password(request_rec *r, const char *user,
	const char *password) {
	dir_config *conf = ap_get_module_config(r->per_dir_config,
		&authn_pam_module);
	struct pam_conv pam_conversation = { conv, (void *) password };
	pam_handle_t *pamh;

	int pam_status = pam_start(conf->service_name, user, &pam_conversation,
		&pamh);
	if (pam_status != PAM_SUCCESS) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"PAM: Could not start service \"%s\"",
			conf->service_name);
		return AUTH_GENERAL_ERROR;
	}

	pam_status = pam_fail_delay(pamh, conf->fail_delay);
	if (pam_status != PAM_SUCCESS) {
		pam_end(pamh, pam_status);
		return AUTH_GENERAL_ERROR;
	}

	pam_status = pam_set_item(pamh, PAM_RHOST,
		ap_get_remote_host(r->connection, r->per_dir_config,
		REMOTE_NAME, NULL));
	if (pam_status != PAM_SUCCESS) {
		pam_end(pamh, pam_status);
		return AUTH_GENERAL_ERROR;
	}

	pam_status = pam_authenticate(pamh, PAM_DISALLOW_NULL_AUTHTOK);
	if (pam_status != PAM_SUCCESS) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"PAM: Authentication error for user \"%s\": %s", user,
			pam_strerror(pamh, pam_status));
		pam_end(pamh, pam_status);
		switch (pam_status) {
		case PAM_AUTH_ERR:
			return AUTH_DENIED;
		case PAM_USER_UNKNOWN:
			return AUTH_USER_NOT_FOUND;
		default:
			return AUTH_GENERAL_ERROR;
		}
	}

	pam_status = pam_acct_mgmt(pamh, PAM_DISALLOW_NULL_AUTHTOK);
	if (pam_status != PAM_SUCCESS) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"PAM: Account error for user \"%s\": %s", user,
			pam_strerror(pamh, pam_status));
		pam_end(pamh, pam_status);
		return AUTH_DENIED;
	}

	return pam_end(pamh, PAM_SUCCESS) == PAM_SUCCESS ? AUTH_GRANTED
		: AUTH_GENERAL_ERROR;
}

static int conv(int num_msg, const struct pam_message **msg,
	struct pam_response **resp, void *appdata_ptr) {
	if (!num_msg || !msg || !resp || !appdata_ptr)
		return PAM_CONV_ERR;

	if (!num_msg) {
		*resp = NULL;
		return PAM_SUCCESS;
	}

	int i;
	for (i = 0; i < num_msg; i++)
		if (msg[i]->msg_style != PAM_PROMPT_ECHO_OFF)
			return PAM_CONV_ERR;

	const char *password = appdata_ptr;

	struct pam_response *response = calloc(num_msg,
		sizeof(struct pam_response));
	if (!response)
		return PAM_BUF_ERR;

	for (i = 0; i < num_msg; i++) {
		char *s = strdup(password);
		if (!s) {
			for (i--; i >= 0; i--)
				free(response[i].resp);
			free(response);
			return PAM_BUF_ERR;
		}
		response[i].resp = s;
	}

	*resp = response;
	return PAM_SUCCESS;
}

static void *create_dir_config(apr_pool_t *p,
	__attribute__((unused)) char *dirspec) {
	dir_config *new = apr_palloc(p, sizeof(dir_config));
	new->fail_delay = 0;
	new->service_name = "apache2";
	return new;
}

static authn_status get_realm_hash(__attribute__((unused)) request_rec *r,
	__attribute__((unused)) const char *user,
	__attribute__((unused)) const char *realm,
	__attribute__((unused)) char **rethash) {
	return AUTH_USER_NOT_FOUND;
}

static void register_hooks(apr_pool_t *p) {
	ap_register_provider(p, AUTHN_PROVIDER_GROUP, "pam", "0",
		&authn_pam_provider);
}
